/*
                                A Sample Filter

   The  following  sample logs each message to a separate temporary file,
   adds  a  recipient  given  with  the -a flag, and rejects a disallowed
   recipient  address given with the -r flag. It recognizes the following
   options:

     -p port The port through which the MTA will connect to the filter.
     -t sec  The timeout value.
     -r addr A recipient to reject.
     -a addr A recipient to add.
     _________________________________________________________________
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>
#include <syslog.h>

#include "libmilter/mfapi.h"
#ifdef DEBUG
#include <mcheck.h>
#endif

#ifndef bool
# define bool   int
# define TRUE   1
# define FALSE  0
#endif /* ! bool */


struct mlfiPriv
{
        char    *mlfi_fname;
        char    *mlfi_connectfrom;
        char    *mlfi_helofrom;
        FILE    *mlfi_fp;
};

#define MLFIPRIV        ((struct mlfiPriv *) smfi_getpriv(ctx))

extern sfsistat         mlfi_cleanup(SMFICTX *, bool);

/* recipients to add and reject (set with -a and -r options) */
char *add = NULL;
char *reject = NULL;
#if _FFR_MILTER_REWRITE
char *do_rewrite = NULL;
#endif /* _FFR_MILTER_REWRITE */
#if _FFR_MILTER_SM_CLASS
char *do_class_lookup = NULL;
#endif /* _FFR_MILTER_SM_CLASS */
#if _FFR_MILTER_SM_MAP
char *do_map_lookup = NULL;
#endif /* _FFR_MILTER_SM_MAP */
int rewrite_flags = SMFRW_NONE;
int test_spike = 0;
char *Name = NULL;
char *String = NULL;

sfsistat
mlfi_connect(ctx, hostname, hostaddr)
         SMFICTX *ctx;
         char *hostname;
         _SOCK_ADDR *hostaddr;
{
	printf("%s: mlfi_connect() hostname=%s\n",
			smfi_getsymval(ctx,"i"),
			hostname);
#if _FFR_MILTER_SM_CLASS

	if (do_class_lookup)
	{
		char **pp = NULL;
		char *p[32];
		int retstat = 0;
		int i = 0;
		int class_options;
		int been_here = 0;
		char *lname;


		if (*do_class_lookup)
			lname = do_class_lookup;
		else
			lname = "w";
do_class_dump:
		pp = NULL;
		if(smfi_sm_vclass(ctx,lname,&pp,&retstat, rewrite_flags | SMFRW_DUMP, NULL) == MI_SUCCESS)
		{
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);
			
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			printf("%s: smfi_sm_class(): failed\n",
				smfi_getsymval(ctx,"i"));

		pp = NULL;
		if(!been_here &&
		   (smfi_sm_vclass(ctx, lname, &pp, &retstat, rewrite_flags | SMFRW_ADD, 
				  "localhost", "glookus", NULL) == MI_SUCCESS)
		  )
		{
			been_here = 1;
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);

			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
			goto do_class_dump;
		}
		else if (!been_here)
			printf("%s: smfi_sm_class(): failed\n",
				smfi_getsymval(ctx,"i"));

		p[0] = "localhost";
		p[1] = "nameserver3";
		p[2] = "gookus";
		p[3] = NULL;

		pp = p;

		if (smfi_sm_class(ctx, lname, &pp, &retstat, rewrite_flags) == MI_SUCCESS)
		{
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			return SMFIS_CONTINUE;
		
		pp = p;
		if (smfi_sm_vclass(ctx, lname, &pp, &retstat, rewrite_flags, String, NULL)
		    == MI_SUCCESS)
		{
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			return SMFIS_CONTINUE;

		pp = p;
		if (smfi_sm_vclass(ctx, lname, &pp, &retstat, rewrite_flags | SMFRW_ALLMATCH, String, NULL)
		    == MI_SUCCESS)
		{
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			return SMFIS_CONTINUE;

		pp = p;
		if (smfi_sm_vclass(ctx, lname, &pp, &retstat, rewrite_flags | SMFRW_ALLMATCH | SMFRW_MULTIPLE_CONT, 
					String, NULL)
		    == MI_SUCCESS)
		{
			printf("%s: smfi_sm_class(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					retstat, pp);
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_class(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			return SMFIS_CONTINUE;
	}
#endif /* _FFR_MILTER_SM_CLASS */
			
					


        /* continue processing */

        return SMFIS_CONTINUE;
}

sfsistat
mlfi_helo(ctx, helohost)
         SMFICTX *ctx;
         char *helohost;
{
	printf("%s: mlfi_helo() helohost=%s\n",
			smfi_getsymval(ctx,"i"),
			helohost);
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_envfrom(ctx, argv)
         SMFICTX *ctx;
         char **argv;
{
	int i = 0;
	
	printf("%s: mlfi_envfrom() ",smfi_getsymval(ctx,"i"));
	while(argv[i])
	{
		printf("argv[%i]=%s ",i,argv[i]); 
		i++;
	}
	printf("\n");
	
        /* continue processing */
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_envrcpt(ctx, argv)
         SMFICTX *ctx;
         char **argv;
{
	int i = 0;
	sfsistat retstat = SMFIS_CONTINUE;
	
	printf("%s: mlfi_envrcpt() ",smfi_getsymval(ctx,"i"));
	while(argv[i])
	{
		printf("argv[%i]=%s ",i,argv[i]); 
		i++;
	}
	printf("\n");
	
	printf("%s: mlfi_envrcpt() macros: "
	       "rcpt_mailer==\"%s\" "
	       "rcpt_host==\"%s\" "
	       "rcpt_addr==\"%s\" "
	       "addr_type==\"%s\" "
	       "\n",
	       smfi_getsymval(ctx,"i"),
	       smfi_getsymval(ctx,"{rcpt_mailer}"),
	       smfi_getsymval(ctx,"{rcpt_host}"),
	       smfi_getsymval(ctx,"{rcpt_addr}"),
	       smfi_getsymval(ctx,"{addr_type")
	      );

#if _FFR_MILTER_REWRITE
	if (do_rewrite) 
	{
		char *p[5];
		char **pp = p;
		int ret_status;
		int ret;
		int i = 0;
		int vrewrite = 0;
		int multiple_rw = 0;

		pp[0] = argv[0];
		pp[1] = argv[1];
		pp[2] = argv[2];
		pp[3] = NULL;
		
		ret = smfi_rewrite(ctx, "spamFriend", &pp, &ret_status, rewrite_flags);
	proc_rewrite:
		if(ret == MI_SUCCESS)
		{
			printf("%s: smfi_%sewrite() returned: result code of %i\n",
					smfi_getsymval(ctx,"i"),
					(vrewrite) ? "vr" : "r",  
					ret_status);
			i = 0;
			while(pp[i])
			{
				printf("%s: smfi_%sewrite() returned: string \"%s\"\n",
					smfi_getsymval(ctx,"i"),
					(vrewrite) ? "vr" : "r",  
					pp[i]);
				
				free(pp[i]);
				i++;
						
			}
			printf("%s: smfi_%sewrite() stored strings in 0x%p, ours were in 0x%p\n",
					smfi_getsymval(ctx,"i"),
					(vrewrite) ? "vr" : "r",  
					p, pp);
			free(pp);
			if(ret_status < 0)
			{
				printf("%s: mlfi_envrcpt() returned SMFIS_REJECT\n",
					smfi_getsymval(ctx,"i"));

				retstat = SMFIS_REJECT;
			}
			if(ret_status > 0)
			{
				printf("%s: mlfi_envrcpt() returned SMFIS_ACCEPT\n",
					smfi_getsymval(ctx,"i"));
				retstat = SMFIS_ACCEPT;
			}
		}
		else
			printf("%s: smfi_%sewrite() returned %i\n",
					smfi_getsymval(ctx,"i"),
					(vrewrite) ? "vr" : "r",  
					ret);

		pp = p;
		if (vrewrite == 0)
		{
			ret = smfi_vrewrite(ctx, "spamFriend", &pp, &ret_status, rewrite_flags, p[0], p[1]);
			vrewrite = 1;
			goto proc_rewrite;
		}

		if (multiple_rw == 0 && (SMFRW_MULTIPLE & rewrite_flags))
		{
			ret = smfi_vrewrite(ctx, "3,2,0", &pp, &ret_status, rewrite_flags, p[0], p[1]);
			multiple_rw = 1;
			goto proc_rewrite;
		}
		if (multiple_rw < 2 && *do_rewrite)
		{
			pp = p;
			ret = smfi_vrewrite(ctx, do_rewrite, &pp, &ret_status, rewrite_flags, String, NULL);
			multiple_rw++;
			goto proc_rewrite;
		}
	}
	else
		printf("%s: not doing rewrite\n",
					smfi_getsymval(ctx,"i"));
#endif /* _FFR_MILTER_REWRITE */

#if _FFR_MILTER_SM_MAP

	if (do_map_lookup)
	{
		char *p[32];
		char **pp = p;
		char * mapname = do_map_lookup;
		int ret_status;
		int ret;

		if (!*do_map_lookup)
			mapname = "bestmx"; 
		else
			mapname = do_map_lookup;

		p[0] = NULL;
		if((ret = smfi_sm_vmap(ctx, mapname, &pp, &ret_status, rewrite_flags | SMFRW_MULTIPLE_CONT, String, NULL)) ==
				MI_SUCCESS)
		{
			printf("%s: smfi_sm_vmap(): returned status %d and strings 0x%p\n",
					smfi_getsymval(ctx,"i"),
					ret_status, pp);
			
			for (i = 0; pp && pp[i]; i++)
			{
				printf("%s: smfi_sm_vmap(): string %d : \"%s\"\n",
					smfi_getsymval(ctx,"i"),i,
					pp[i]);
				free(pp[i]);
			}
			if (pp)
				free(pp);
		}
		else
			printf("%s: smfi_sm_vmap(): call failed, status %d \n",
					smfi_getsymval(ctx,"i"),
					ret);

	}
	else
		printf("%s: not doing map\n",
					smfi_getsymval(ctx,"i"));

#endif /* _FFR_MILTER_SM_MAP */
	
        /* continue processing */
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_header(ctx, headerf, headerv)
         SMFICTX *ctx;
         char *headerf;
         unsigned char *headerv;
{
	int i = 0;
	
	printf("%s: mlfi_header() headerf=%s headerv=%s\n",
			smfi_getsymval(ctx,"i"),
			headerf,
			headerv);
	
	if (test_spike)
	{
		printf("%s: mlfi_eom() testing spike\n",
				smfi_getsymval(ctx,"i"));
		return SMFIS_REJECT;
	}
	
        /* continue processing */
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_eoh(ctx)
         SMFICTX *ctx;
{
	printf("%s: mlfi_eoh()\n",
			smfi_getsymval(ctx,"i"));
        /* continue processing */
	if (test_spike)
	{
		printf("%s: mlfi_eom() testing spike\n",
				smfi_getsymval(ctx,"i"));
		return SMFIS_REJECT;
	}
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_body(ctx, bodyp, bodylen)
         SMFICTX *ctx;
         unsigned char *bodyp;
         size_t bodylen;
{
	printf("%s: mlfi_eoh() bodylen=%i BODY\n%s\n",
			smfi_getsymval(ctx,"i"),
			bodylen,
			bodyp
			);
	if (test_spike)
	{
		printf("%s: mlfi_eom() testing spike\n",
				smfi_getsymval(ctx,"i"));
		return SMFIS_REJECT;
	}
        /* continue processing */
        return SMFIS_CONTINUE;
}

sfsistat
mlfi_eom(ctx)
         SMFICTX *ctx;
{
	printf("%s: mlfi_eom()\n",
			smfi_getsymval(ctx,"i"));
	if (test_spike)
	{
		printf("%s: mlfi_eom() testing spike\n",
				smfi_getsymval(ctx,"i"));
		return SMFIS_REJECT;
	}
	return SMFIS_ACCEPT;
}

sfsistat
mlfi_abort(ctx)
         SMFICTX *ctx;
{
	printf("%s: mlfi_abort()\n",
			smfi_getsymval(ctx,"i"));
	return SMFIS_CONTINUE;
}

sfsistat
mlfi_close(ctx)
         SMFICTX *ctx;
{
	printf("%s: mlfi_close()\n",
			smfi_getsymval(ctx,"i"));
        return SMFIS_CONTINUE;
}

struct smfiDesc smfilter =
{
        "Verbose-Milter", /* filter name */
        SMFI_VERSION,   /* version code -- do not change */
        SMFIF_ADDHDRS,  /* flags */
        mlfi_connect,       /* connection info filter */
        mlfi_helo,  /* SMTP HELO command filter */
        mlfi_envfrom,       /* envelope sender filter */
        mlfi_envrcpt,       /* envelope recipient filter */
        mlfi_header,        /* header filter */
        mlfi_eoh,   /* end of header */
        mlfi_body,  /* body block filter */
        mlfi_eom,   /* end of message */
        mlfi_abort, /* message aborted */
        mlfi_close, /* connection cleanup */
};

static void
usage(prog)
        char *prog;
{
        fprintf(stderr,
                "Usage: %s [-d dbg_lvl] -[INAEDPCSTfmg] [-c Name] [-w Name] [-M Name] [-s String] -p socket-addr [-t timeout]\n"
		"	-I -- Invert status that MTA returns\n"
		"	-N -- Dont return any strings from MTA calls\n"
		"	-A -- class ALLMATCH\n"
		"	-E -- sendmail macro expansion\n"
		"	-D -- convert sendmail dollar symbols\n"
		"	-P -- perform sendmail prescan\n"
		"	-d dbg_lvl -- numeric value\n"
		"	-g -- turn on eagain\n"
		"	-w Name -- turn on rewriting, use name for some operations\n"
		"	-M Name -- turn on map lookup, use name for some operations\n"
		"	-m -- turn on multiple names\n"
		"	-f -- turn on unfiltered rewriting\n"
		"	-C -- turn on CONDELSE string seperator\n"
		"	-c Name -- turn on class lookups, use name for some operations\n"
		"	-S -- Our rejecions may be spiked\n"
		"	-T -- Turn on SMFIF_TRYREJCT\n"
		"	-s String -- used as parameter to MTA calls\n"
                ,prog);
}

int
main(argc, argv)
         int argc;
         char **argv;
{
        bool setconn = FALSE;
        int c;
        const char *args = "AEDPTSCM:c:fw:gs:md:p:t:h";
        extern char *optarg;

#ifdef DEBUG
	mtrace();
#endif

        /* Process command line options */
        while ((c = getopt(argc, argv, args)) != -1)
        {
                switch (c)
                {
		  case 's':
			  String = optarg;
			  break;
		  case 'd':
			smfi_setdbg(atoi(optarg));
			break;
                  case 'p':
                        if (optarg == NULL || *optarg == '\0')
                        {
                                (void) fprintf(stderr, "Illegal conn: %s\n",
                                               optarg);
                                exit(EX_USAGE);
                        }
                        if (smfi_setconn(optarg) == MI_FAILURE)
                        {
                                (void) fprintf(stderr,
                                               "smfi_setconn failed\n");
                                exit(EX_SOFTWARE);
                        }

                        /*
                        **  If we're using a local socket, make sure it
                        **  doesn't already exist.  Don't ever run this
                        **  code as root!!
                        */

                        if (strncasecmp(optarg, "unix:", 5) == 0)
                                unlink(optarg + 5);
                        else if (strncasecmp(optarg, "local:", 6) == 0)
                                unlink(optarg + 6);
                        setconn = TRUE;
                        break;

                  case 't':
                        if (optarg == NULL || *optarg == '\0')
                        {
                                (void) fprintf(stderr, "Illegal timeout: %s\n",
                                               optarg);
                                exit(EX_USAGE);
                        }
                        if (smfi_settimeout(atoi(optarg)) == MI_FAILURE)
                        {
                                (void) fprintf(stderr,
                                               "smfi_settimeout failed\n");
                                exit(EX_SOFTWARE);
                        }
                        break;
#if _FFR_MILTER_SM_CLASS
		  case 'c':
			do_class_lookup = optarg;
			smfilter.xxfi_flags |= SMFIF_SM_CLASS;
			break;
#else
		case 'c':
			break;
#endif /* _FFR_MILTER_SM_CLASS */
#if _FFR_MILTER_REWRITE
		  case 'w':
			/*Turn on Rewrite/Eagain support*/
			smfilter.xxfi_flags |= SMFIF_REWRITE;
			do_rewrite = optarg;
			break;
#else
		case 'w':
			break;
#endif /* _FFR_MILTER_REWRITE */
#if _FFR_MILTER_SM_MAP
		  case 'M':
			smfilter.xxfi_flags |= SMFIF_SM_MAP;
			do_map_lookup = optarg;
			break;
#else
		  case 'M':
			break;
#endif /* _FFR_MILTER_SM_MAP */
#if _FFR_MILTER_REWRITE || _FFR_MILTER_SM_CLASS || _FFR_MILTER_SM_MAP || _FFR_MILTER_SM_MACRO
		  case 'I':
			rewrite_flags |= SMFRW_NEGMATCH;
			break;
		  case 'N':
			rewrite_flags |= SMFRW_NORETSTR;
			break;
		  case 'f':
			/* Set rewrite unfiltered flag*/
			rewrite_flags |= SMFRW_UNFILTER;
		  	break;
		  case 'C':
			/* turn on CONDELSE rewrite flag*/
			rewrite_flags |= SMFRW_CONDELSE;
			break;
		  case 'm':
			/* turn on multiple rulesets calls*/
			rewrite_flags |= SMFRW_MULTIPLE;
			break;
		  case 'P':
			/* turn on prescan */
			rewrite_flags |= SMFRW_PRESCAN;
			break;
		  case 'D':
			/* turn on dollars */
			rewrite_flags |= SMFRW_DOLLARS;
			break;
		  case 'E':
			/* turn on expansion */
			rewrite_flags |= SMFRW_EXPAND;
			break;
		  case 'A':
			/* turn on ALLMTACH */
			rewrite_flags |= SMFRW_ALLMATCH;
			break;
#endif /* _FFR_MILTER_REWRITE || _FFR_MILTER_SM_CLASS || _FFR_MILTER_SM_MAP || _FFR_MILTER_SM_MACRO */
#if _FFR_MILTER_EAGAIN
		  case 'g':
			smfilter.xxfi_flags |= SMFIF_EAGAIN; 
			break;
#else
		case 'g':
			break;
#endif /* _FFR_MILTER_EAGAIN */
#if _FFR_MILTER_RWSETS_SPIKE
		  case 'S':
			smfilter.xxfi_flags |= SMFIF_TRYREJECT;
			break;
		  case 'T':
			test_spike = 1;
			break;
#else
		case 'S':
		case 'T':
			break;
#endif /* _FFR_MILTER_RWSETS_SPIKE */
                  case 'h':
                  default:
                        usage(argv[0]);
                        exit(EX_USAGE);
                }
        }
        if (!setconn)
        {
                fprintf(stderr, "%s: Missing required -p argument\n", argv[0]);
                usage(argv[0]);
                exit(EX_USAGE);
        }
        if (smfi_register(smfilter) == MI_FAILURE)
        {
                fprintf(stderr, "smfi_register failed\n");
                exit(EX_UNAVAILABLE);
        }
	openlog(smfilter.xxfi_name,LOG_PID,LOG_MAIL);
        return smfi_main();
}

/* eof */

/*
     _________________________________________________________________

   Copyright  (c)  2000-2003 Sendmail, Inc. and its suppliers. All rights
   reserved.
   By using this file, you agree to the terms and conditions set forth in
   the LICENSE.

 */
